@@ -241,9 +241,8 @@ Geometry::Class_info *Geometry::find_class(const char *name, size_t len)
   {
     if (*cur_rt &&
 	((*cur_rt)->m_name.length == len) &&
-	(my_strnncoll(&my_charset_latin1,
-		      (const uchar*) (*cur_rt)->m_name.str, len,
-		      (const uchar*) name, len) == 0))
+	(my_charset_latin1.strnncoll((*cur_rt)->m_name.str, len,
+		                     name, len) == 0))
       return *cur_rt;
   }
   return 0;
@@ -277,6 +276,33 @@ Geometry *Geometry::construct(Geometry_buffer *buffer,
 }
 
 
+uint Geometry::get_key_image_itMBR(LEX_CSTRING &src, uchar *buff, uint length)
+{
+  const char *dummy;
+  MBR mbr;
+  Geometry_buffer buffer;
+  Geometry *gobj;
+  const uint image_length= SIZEOF_STORED_DOUBLE*4;
+
+  if (src.length < SRID_SIZE)
+  {
+    bzero(buff, image_length);
+    return image_length;
+  }
+  gobj= Geometry::construct(&buffer, (char*) src.str, (uint32) src.length);
+  if (!gobj || gobj->get_mbr(&mbr, &dummy))
+    bzero(buff, image_length);
+  else
+  {
+    float8store(buff,    mbr.xmin);
+    float8store(buff+8,  mbr.xmax);
+    float8store(buff+16, mbr.ymin);
+    float8store(buff+24, mbr.ymax);
+  }
+  return image_length;
+}
+
+
 Geometry *Geometry::create_from_wkt(Geometry_buffer *buffer,
 				    Gis_read_stream *trs, String *wkt,
 				    bool init_stream)
@@ -520,8 +546,8 @@ Geometry *Geometry::create_from_json(Geometry_buffer *buffer,
             goto create_geom;
         }
         else if (je->value_len == feature_coll_type_len &&
-            my_strnncoll(&my_charset_latin1, je->value, je->value_len,
-		         feature_coll_type, feature_coll_type_len) == 0)
+            my_charset_latin1.strnncoll(je->value, je->value_len,
+		                        feature_coll_type, feature_coll_type_len) == 0)
         {
           /*
             'FeatureCollection' type found. Handle the 'Featurecollection'/'features'
@@ -532,8 +558,8 @@ Geometry *Geometry::create_from_json(Geometry_buffer *buffer,
           fcoll_type_found= 1;
         }
         else if (je->value_len == feature_type_len &&
-                 my_strnncoll(&my_charset_latin1, je->value, je->value_len,
-		              feature_type, feature_type_len) == 0)
+                 my_charset_latin1.strnncoll(je->value, je->value_len,
+		                             feature_type, feature_type_len) == 0)
         {
           if (geometry_start)
             goto handle_geometry_key;
@@ -932,8 +958,7 @@ static int read_point_from_json(json_engine_t *je, bool er_on_3D,
       goto bad_coordinates;
 
     d= (n_coord == 0) ? x : ((n_coord == 1) ? y : &tmp);
-    *d= my_strntod(je->s.cs, (char *) je->value,
-                   je->value_len, &endptr, &err);
+    *d= je->s.cs->strntod((char *) je->value, je->value_len, &endptr, &err);
     if (err)
       goto bad_coordinates;
     n_coord++;
@@ -1071,9 +1096,10 @@ double Gis_point::calculate_haversine(const Geometry *g,
     point_temp[point_size-1]= '\0';
     Geometry_buffer gbuff;
     Geometry *gg= Geometry::construct(&gbuff, point_temp, point_size-1);
-    if (!gg || static_cast<Gis_point *>(gg)->get_xy_radian(&x2r, &y2r))
+    DBUG_ASSERT(gg);
+    if (static_cast<Gis_point *>(gg)->get_xy_radian(&x2r, &y2r))
     {
-      *error= 2;
+      DBUG_ASSERT(0);
       return -1;
     }
   }
@@ -1081,16 +1107,15 @@ double Gis_point::calculate_haversine(const Geometry *g,
   {
     if (static_cast<const Gis_point *>(g)->get_xy_radian(&x2r, &y2r))
     {
-      *error= 2;
+      DBUG_ASSERT(0);
       return -1;
     }
   }
   if (this->get_xy_radian(&x1r, &y1r))
   {
-    *error= 2;
+    DBUG_ASSERT(0);
     return -1;
   }
-  //
   // Check boundary conditions: longitude[-180,180]
   if (!((x2r >= -M_PI && x2r <= M_PI) && (x1r >= -M_PI && x1r <= M_PI)))
   {
@@ -1143,20 +1168,15 @@ int Gis_point::spherical_distance_multipoints(Geometry *g, const double r,
   {
     Geometry_buffer buff_temp;
     Geometry *temp;
-    const char *pt_ptr= g->get_data_ptr()+
-      4+WKB_HEADER_SIZE*i + POINT_DATA_SIZE*(i-1);
 
     // First 4 bytes are handled already, make sure to create a Point
     memset(s + 4, Geometry::wkb_point, 1);
-    if (g->no_data(pt_ptr, POINT_DATA_SIZE))
-      return 1;
-
     memcpy(s + 5, g->get_data_ptr() + 5, 4);
-    memcpy(s + 4 + WKB_HEADER_SIZE, pt_ptr, POINT_DATA_SIZE);
+    memcpy(s + 4 + WKB_HEADER_SIZE, g->get_data_ptr() + 4 + WKB_HEADER_SIZE*i +\
+                                    POINT_DATA_SIZE*(i-1), POINT_DATA_SIZE);
     s[len-1]= '\0';
     temp= Geometry::construct(&buff_temp, s, len);
-    if (!temp)
-      return 1;
+    DBUG_ASSERT(temp);
     temp_res= this->calculate_haversine(temp, r, err);
     if (res > temp_res)
       res= temp_res;
@@ -2333,18 +2353,14 @@ int Gis_multi_point::spherical_distance_multipoints(Geometry *g, const double r,
     Geometry *temp;
     double temp_res= 0.0;
     char s[len];
-    const char *pt_ptr= get_data_ptr()+
-      4+WKB_HEADER_SIZE*i + POINT_DATA_SIZE*(i-1);
     // First 4 bytes are handled already, make sure to create a Point
     memset(s + 4, Geometry::wkb_point, 1);
-    if (no_data(pt_ptr, POINT_DATA_SIZE))
-      return 1;
     memcpy(s + 5, this->get_data_ptr() + 5, 4);
-    memcpy(s + 4 + WKB_HEADER_SIZE, pt_ptr, POINT_DATA_SIZE);
+    memcpy(s + 4 + WKB_HEADER_SIZE, this->get_data_ptr() + 4 + WKB_HEADER_SIZE*i +\
+                                    POINT_DATA_SIZE*(i-1), POINT_DATA_SIZE);
     s[len-1]= '\0';
     temp= Geometry::construct(&buff_temp, s, len);
-    if (!temp)
-      return 1;
+    DBUG_ASSERT(temp);
     // Optimization for single Multipoint
     if (num_of_points2 == 1)
     {
@@ -2356,18 +2372,14 @@ int Gis_multi_point::spherical_distance_multipoints(Geometry *g, const double r,
       Geometry_buffer buff_temp2;
       Geometry *temp2;
       char s2[len];
-      const char *pt_ptr= g->get_data_ptr()+
-        4+WKB_HEADER_SIZE*j + POINT_DATA_SIZE*(j-1);
       // First 4 bytes are handled already, make sure to create a Point
       memset(s2 + 4, Geometry::wkb_point, 1);
-      if (g->no_data(pt_ptr, POINT_DATA_SIZE))
-        return 1;
       memcpy(s2 + 5, g->get_data_ptr() + 5, 4);
-      memcpy(s2 + 4 + WKB_HEADER_SIZE, pt_ptr, POINT_DATA_SIZE);
+      memcpy(s2 + 4 + WKB_HEADER_SIZE, g->get_data_ptr() + 4 + WKB_HEADER_SIZE*j +\
+                                       POINT_DATA_SIZE*(j-1), POINT_DATA_SIZE);
       s2[len-1]= '\0';
       temp2= Geometry::construct(&buff_temp2, s2, len);
-      if (!temp2)
-        return 1;
+      DBUG_ASSERT(temp2);
       temp_res= static_cast<Gis_point *>(temp)->calculate_haversine(temp2, r, err);
       if (res > temp_res)
         res= temp_res;
@@ -3270,9 +3282,7 @@ bool Gis_geometry_collection::init_from_wkt(Gis_read_stream *trs, String *wkb)
       return 1;
 
     if (next_word.length != 5 ||
-	(my_strnncoll(&my_charset_latin1,
-		      (const uchar*) "empty", 5,
-		      (const uchar*) next_word.str, 5) != 0))
+	(my_charset_latin1.strnncoll("empty", 5, next_word.str, 5) != 0))
     {
       for (;;)
       {
