@@ -1,7 +1,7 @@
 /***********************************************************************/
-/*  Name: ODBCONN.CPP  Version 2.3                                     */
+/*  Name: ODBCONN.CPP  Version 2.4                                     */
 /*                                                                     */
-/*  (C) Copyright to the author Olivier BERTRAND          1998-2017    */
+/*  (C) Copyright to the author Olivier BERTRAND          1998-2021    */
 /*                                                                     */
 /*  This file contains the ODBC connection classes functions.          */
 /***********************************************************************/
@@ -1419,7 +1419,7 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
   PGLOBAL& g = m_G;
   void    *buffer;
   bool     b;
-  UWORD    n;
+  UWORD    n, k;
   SWORD    len, tp, ncol = 0;
   ODBCCOL *colp;
   RETCODE  rc;
@@ -1486,15 +1486,16 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
     } else {
       do {
         rc = SQLExecDirect(hstmt, (PUCHAR)sql, SQL_NTS);
-        } while (rc == SQL_STILL_EXECUTING);
+      } while (rc == SQL_STILL_EXECUTING);
 
       if (!Check(rc))
         ThrowDBX(rc, "SQLExecDirect", hstmt);
 
       do {
         rc = SQLNumResultCols(hstmt, &ncol);
-        } while (rc == SQL_STILL_EXECUTING);
+      } while (rc == SQL_STILL_EXECUTING);
 
+      k = 0;    // used for column number
     } // endif Srcdef
 
     for (n = 0, colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
@@ -1506,7 +1507,7 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
       ThrowDBX(MSG(COL_NUM_MISM));
 
     // Now bind the column buffers
-    for (n = 1, colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
+    for (colp = tocols; colp; colp = (PODBCCOL)colp->GetNext())
       if (!colp->IsSpecial()) {
         buffer = colp->GetBuffer(m_RowsetSize);
         len = colp->GetBuflen();
@@ -1516,19 +1517,23 @@ int ODBConn::ExecDirectSQL(char *sql, ODBCCOL *tocols)
           sprintf(m_G->Message, MSG(INV_COLUMN_TYPE),
                   colp->GetResultType(), SVP(colp->GetName()));
           ThrowDBX(m_G->Message);
-          } // endif tp
+        } // endif tp
+
+        if (m_Tdb->Srcdef)
+          k = colp->GetIndex();
+        else
+          k++;
 
         if (trace(1))
           htrc("Binding col=%u type=%d buf=%p len=%d slen=%p\n",
-                  n, tp, buffer, len, colp->GetStrLen());
+                  k, tp, buffer, len, colp->GetStrLen());
 
-        rc = SQLBindCol(hstmt, n, tp, buffer, len, colp->GetStrLen());
+        rc = SQLBindCol(hstmt, k, tp, buffer, len, colp->GetStrLen());
 
         if (!Check(rc))
           ThrowDBX(rc, "SQLBindCol", hstmt);
 
-        n++;
-        } // endif pcol
+      } // endif colp
 
   } catch(DBX *x) {
     if (trace(1))
